;-----------------------------------------------------------------------------
;                       MBR for Partition Installation
;-----------------------------------------------------------------------------
; This MBR code should be installed when the bootmanager is installed in a
; partition. The MBR code will search all the drives, all bootmanager parti-
; tions until it find the right bootmanager partition.
;
; We assume the following in this code:
; - Sector size is 512 bytes.
; - 16 directory entries per sector.



;-----------------------------------------------------------------------------
; This is a list of variables defined elsewhere, and some defined here but
; used elsewhere.

;extern  BootmanagerSegment      ; Linear location in memory where the boot-
;extern  BootmanagerOffset       ; manager should be loaded.


MBR_Addr:       equ     07C00h
MBR_RunAddr:    equ     07B00h
BS_Addr:        equ     07E00h
Root_Addr:      equ     08000h
FAT_Addr:       equ     0C000h

%include        "INCLUDE\MBR.INC"
%include        "INCLUDE\FAT16.INC"

;-----------------------------------------------------------------------------
; Main Code Segment
;-----------------------------------------------------------------------------
; This part will contain the main code segment.

;segment BootPartitionMBR
        org     07C00h



;-----------------------------------------------------------------------------
; Initialize enviroment
;-----------------------------------------------------------------------------
; This code will set the registers to a known state. The code should always
; run at linear address 00007C00h. And since we are in real mode, we'll load
; DS with 0000h and use the base 7C00h we doing memory accesses.

        xor     ax, ax
        mov     ds, ax
        mov     ss, ax

; The drive number of the drive we booted from should be in DL, that one we'll
; save for later use.
        mov     byte [MBR_RunAddr+MaxCHS.BootDrive], dl



        push    es
        mov     ah, 008h
        ; dl should already contain the drive number.
        int     013h
        pop     es
        jc      near FatalError

        inc     dh
        mov     byte [MBR_RunAddr+MaxCHS.Head], dh

        and     cl, 03Fh
        mov     byte [MBR_RunAddr+MaxCHS.Sector], cl



; Load boot sector.
        mov     eax, [MBR_Addr+MBR.PartitionEntry1+PartitionTableEntry.RelativeSector]      ; LBA of boot sector.
        mov     bx, BS_Addr             ; Memory location where to put data.
        mov     cx, 00001h              ; Number of sectors to read.
        call    LoadSectors



; Now we got the LBA address of our partition. Next we are going to calculate
; the LBA address of our first FAT. That should be:
; Start + Hidden Sector
        movzx   eax, word [BS_Addr+FAT16.ReservedSectors]
        add     eax, [MBR_Addr+MBR.PartitionEntry1+PartitionTableEntry.RelativeSector]
        mov     dword [MBR_RunAddr+LBA_Addr.FAT], eax

; And then the LBA address of the root directory.
        movzx   eax, word [BS_Addr+FAT16.SectorsPerFAT]
        movzx   cx, byte [BS_Addr+FAT16.NumberOfFATs]
        mul     cx
        add     eax, [MBR_RunAddr+LBA_Addr.FAT]
        mov     dword [MBR_RunAddr+LBA_Addr.RootDir], eax      ; LBA of boot sector.

; And then divide the number of root entries with the number of entries per
; sector, to get the number of sectors the root directory takes.
        movzx   eax, word [BS_Addr+FAT16.RootEntries]
        shr     ax, 004h
        push    ax
        add     eax, [MBR_RunAddr+LBA_Addr.RootDir]
        mov     dword [MBR_RunAddr+LBA_Addr.Data], eax

; Load Root Directory.
        mov     eax, [MBR_RunAddr+LBA_Addr.RootDir]          ; LBA of boot sector.
;        mov     bx, Root_Addr           ; Memory location where to put data.
        pop     cx                      ; Number of sectors to read.
        call    LoadSectors



; Search root directory.
SearchRootDirectory:
        mov     di, Root_Addr
        mov     cx, word [BS_Addr+FAT16.RootEntries]
.Loop1:
        push    cx
        push    di
        mov     si, BootmanagerFileName
        mov     cx, 00Bh
        rep     cmpsb
        je      .FoundIt
        pop     di
        add     di, 020h
        pop     cx
        loop    .Loop1
        jmp     near FatalError

.FoundIt:
        pop     di
        pop     cx



; Load bootmanager.
; All this will calculate number of clusters to load
        mov     eax, dword [edi+DirEntry.Size]

; Calculate the sector count:
        shr     eax, 009h
        inc     eax

; Calculate the cluster count:
        movzx   ecx, byte [BS_Addr+FAT16.SectorsPerCluster]
        xor     edx, edx                ; Clear EDX, top 16 bits should be 0 now.
        div     ecx                     ; Get the number of clusters - 1
        inc     eax                     ; EAX = Number of clusters

; Save the number of clusters to load on the stack.
        push    ax

; Transfer first cluster
        movzx   eax, word [edi+DirEntry.StartClusterLow]

; Calculate the LBA location for the beginning of the cluster.
; AX = Current cluster (16 bit value).
; CX = Contain the number of clusters left.
LoadClusterString:
        pop     cx
        mov     bx, 00000h
.CalculateLBAFromCluster:
        push    cx
        push    eax
        sub     eax, 002h
        movzx   ecx, byte [BS_Addr+FAT16.SectorsPerCluster]
        mul     ecx
        add     eax, [MBR_RunAddr+LBA_Addr.Data]
        push    word 09000h
        pop     es
        mov     cl, byte [BS_Addr+FAT16.SectorsPerCluster] ; Top 24 bits should be 0 now.
        call    LoadSectors
        pop     eax
        pop     cx
        dec     cx
        jz      near .Exit



.LocateNextCluster:
        push    bx
.IsCurrentClusterInFATSection:
        mov     bx, [FAT_Section.FirstEntry]
        cmp     ax, bx
        jb      .UpdateFATSection
        add     bx, FAT_Section.EntriesPerSection
        cmp     ax, bx
        ja      .UpdateFATSection

.ValidateCluster:
        sub     ax, [FAT_Section.FirstEntry]
        mul     word [FAT_Section.BytesPerEntry]
        push    ax
        pop     bx
        add     bx, FAT_Addr
        cmp     word [bx], 00000h
        jz      near .InvalidCluster
        cmp     word [bx], 0FFFFh
        jz      .EndOfFile
        cmp     word [bx], 0FFF8h
        jae     .InvalidCluster
        cmp     word [bx], 0FFF6h
        jae     .BadCluster
        mov     ax, [bx]

        pop     bx

        jmp     .CalculateLBAFromCluster

.UpdateFATSection:
;> AX should contain the cluster number.
        pushad
        shr     ax, 008h
        add     eax, [MBR_RunAddr+LBA_Addr.FAT]

        push    cs
        pop     es
        mov     bx, FAT_Addr
        mov     cx, FAT_Section.SizeInSectors
        call    LoadSectors

        popad
        push    ax
        and     ax, 0FF00h
        mov     word [FAT_Section.FirstEntry], ax
        pop     ax

        jmp     .ValidateCluster



.BadCluster:
.EndOfFile:
.InvalidCluster:
.Exit:
FatalError:
        movzx   ebx, word [edi+DirEntry.StartClusterLow]
        mov     dl, [MBR_RunAddr+MaxCHS.BootDrive]
        mov     cx, Cluster_Address+BIFSInstallation
        
        int3
        jmp     09000h:00200h









LoadSectors:
.PerformIteration:
        push    cx                      ; The number of sectors to read.



; Read or write the given block of sectors, depending on the current state of
; the variables.
        mov     cx, 00004h
.RetryIteration:
        push    cx
        push    eax

; Convert LBA address (in EAX) to Cylinder and sector value (in CX) and head
; value (in DH).
        xor     edx, edx
        movzx   ecx, byte [MBR_RunAddr+MaxCHS.Sector]
        div     ecx
        inc     dl
        movzx   cx, dl

        xor     edx, edx
        push    cx
        movzx     ecx, byte [MBR_RunAddr+MaxCHS.Head]      ; Top 24 bits should be 0
        div     ecx
        pop     cx
        xchg    dl, dh

        xchg    al, ah
        ror     al, 002h
        or      cx, ax

; Load.
        mov     ax, 00201h
        ; BX should contain the memory pointer
        mov     dl, [MBR_RunAddr+MaxCHS.BootDrive]
        int     013h
        pop     eax
        pop     cx
        jnc     .Calculate_RemainingSectors
        loop    .RetryIteration
        pop     cx
        ret



; Calculate remaining sectors.
.Calculate_RemainingSectors:
        inc     eax
        add     bx, 00200h
        pop     cx
        loop    .PerformIteration
        ret




;-----------------------------------------------------------------------------
; Data - Text strings
;-----------------------------------------------------------------------------
; The data used in this code.

BootmanagerFileName:    db      'BOOTMAN SYS'



;-----------------------------------------------------------------------------
; Structures
;-----------------------------------------------------------------------------
; These are some structures which are used by the code.

FAT_Section:
.FirstEntry:            dw      00000h
.BytesPerEntry:         dw      00002h
.SizeInSectors:         equ     001h
.EntriesPerSection:     equ     00100h

struc   LBA_Addr
        .Partition      resd    1
        .FAT            resd    1
        .RootDir        resd    1
        .Data           resd    1
endstruc

struc   MaxCHS
        .Head           resb    1
        .Sector         resb    1
        .BootDrive:     resb    1
endstruc

%include        "INCLUDE\CONFIG.INC"
